/*
 *  pv_main.c
 *  ProteinViewer
 *
 *  Created by Naoto Morikawa on 6/15/06.
 *  Copyright (C) 2006, 2007, 2008 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-04-09 support quit button "Q" in keyboard()
 *  2007-04-09 support CO, CB, CG, CU, and CV in scene_original_cylinder_model2()
 *  2007-04-15 version 0.3 release
 *  2008-05-20 set vctr->z to +0.00001 if "(vctr->x == 0.0) && (vctr->z == 0.0)"
 *             to avoid the "axis = (0,0,0)" error in draw_cylinder()
 */

/* #include <OpenGL/gl.h> */
#include <stdlib.h>
#include <GLUT/glut.h>
#include <stdio.h>
#include <math.h>
#include "pv_main.h"
#include "pv_read_data.h"
#include "pv_primitives.h"

extern float Tcenter[];
extern float Tcenter_save[];
extern float Ssize[];
extern GLuint tetra_table[];
extern GLuint tetra_table2[];
extern GLuint tetra_tablew[];
extern GLuint tetra_table2w[];
extern struct pe_path_of_length_n * npath_table[];

extern struct pe_slt_tile * myinit();

struct pv_macro my_macro;
struct pe_coor_system  my_coor;
struct pe_settings my_settings;
struct pe_slt_tile * slt_tiles;
struct pe_vector * avectors;
struct pe_vector * tvectors;

static GLfloat Xrot = 0.0, Yrot = 0.0;
static GLfloat CamX = 0.0, CamY = 0.0;
static float ScaleFactor = 1.0;

static int pv_blend = ON;     /* Blending OFF/ON */
static int pv_tetra_type = SOLID_TYPE; /* CYLINDER_TYPE, SOLID_TYPE, WIRE_TYPE */
static float pv_zoom_in = 1;
static int pv_rot = OFF;
static int pv_trans = OFF;
static int pv_prt_and_mdl = BOTH_PRT_AND_MDL;

void outer_product(struct pe_vector * pt_rslt, struct pe_vector * pt_vec1,
		   struct pe_vector * pt_vec2)
{
	pt_rslt->x = pt_vec1->y * pt_vec2->z - pt_vec1->z * pt_vec2->y;
	pt_rslt->y = pt_vec1->z * pt_vec2->x - pt_vec1->x * pt_vec2->z;
	pt_rslt->z = pt_vec1->x * pt_vec2->y - pt_vec1->y * pt_vec2->x;
}

void draw_cylinder(struct pe_vector * vctr, int radius)
{
	struct pe_vector cylndr_axis = { 0.0, 1.0, 0.0, 1.0};
	struct pe_vector axis;
	double degree;
	double tmp;
	
	if ( (vctr->x == 0.0) && (vctr->z == 0.0) ) {		// 2008-05-20
		vctr->z = 0.00001;
	}
	 /* A x B = (Ay*Bz - Az*By, Az*Bx - Ax*Bz, Ax*By - Ay*Bx) */
	outer_product(&axis,&cylndr_axis,vctr);
	tmp = cylndr_axis.x * vctr->x + cylndr_axis.y * vctr->y
		+ cylndr_axis.z * vctr->z;
	 /* (p,q) = |p||q|cos(angle) */
	degree = (180.0/PI) * acos(tmp / vctr->length);
	
	glPushMatrix();
	if ( (axis.x==0.0) && (axis.y==0.0) && (axis.z==0.0)){
		//2008-05-20	printf("ERROR in draw_cylinder: axis == (0,0,0)\n");
		printf("ERROR in draw_cylinder: axis == (0,0,0) = cylinder_axis x cur_vector, where\n");
		printf("ERROR in draw_cylinder: cylinder_axis = (%f ,%f, %f)\n",
		       cylndr_axis.x, cylndr_axis.y, cylndr_axis.z);
		printf("ERROR in draw_cylinder: cur_vector = (%f ,%f, %f)\n",
		       vctr->x, vctr->y, vctr->z);
		exit(1);
	}else {
		glRotatef((float)degree, axis.x, axis.y, axis.z);
	}
	glScalef(1.0F, vctr->length, 1.0F);
	switch(radius){
		case  30:
			glCallList(cylinder_30);
			break;
		case  1015:
			glCallList(cylinder_1015);
			break;
		case  2015:
			glCallList(cylinder_2015);
			break;
		case  15:
			glCallList(cylinder_15);
			break;
		case  2:
			glCallList(cylinder_2);
			break;
		case  5:
			glCallList(cylinder_5);
			break;
		case  10:
			glCallList(cylinder_10);
			break;
		case  1010:
			glCallList(cylinder_1010);
			break;
		default:
			glCallList(cylinder_5);
	}
	glPopMatrix();
}

void scene_test(void)
{
	if (0) {
		/* Origin */
		glCallList(sphere_50white);
		glTranslatef(5.0,0.0,0.0);
		glCallList(sphere_green);
		glTranslatef(-5.0,5.0,0.0);
		glCallList(sphere_blue);
		glTranslatef(0.0,-5.0,5.0);
		glCallList(sphere_red);
		glTranslatef(0.0,0.0,-5.0);
	}
	
	/* center */
	glTranslatef (-1.0*Tcenter[0], -1.0*Tcenter[1], -1.0*Tcenter[2]);
	glCallList(sphere_50white);
	glTranslatef(2.0,0.0,0.0);
	glCallList(sphere_green);
	glTranslatef(-2.0,2.0,0.0);
	glCallList(sphere_blue);
	glTranslatef(0.0,-2.0,2.0);
	glCallList(sphere_red);
	glTranslatef(0.0,0.0,-2.0);

/*
	glCallList(sphere_50white);
	glTranslatef(3.0,0.0,0.0);
	glCallList(tetra2_xwzy);
	glTranslatef(2.0,0.0,0.0);
	glCallList(tetra2_xywz);
	glTranslatef(2.0,0.0,0.0);
	glCallList(tetra2_xzyw);
	glTranslatef(2.0,0.0,0.0);
	glCallList(tetra_xwzy);
*/

	if (0) {
		/* npath_table[] */
		int i, p_ct, t_ct, n;
		
		n = (int) my_settings.tiles_per_res;
		for (i = 0; i < 24; i++) {
			for (p_ct = 0; p_ct < my_settings.max_path; p_ct++) {
 				for (t_ct = 0; t_ct < n; t_ct++) {
					draw_cylinder(&(npath_table[i][p_ct]
							.bb_c[t_ct]), 2);
					glTranslatef(
					    npath_table[i][p_ct].bb_c[t_ct].x,
					    npath_table[i][p_ct].bb_c[t_ct].y,
					    npath_table[i][p_ct].bb_c[t_ct].z);
				}
				glCallList(sphere_yellow);
				glTranslatef(-1.0*npath_table[i][p_ct].c.x,
					     -1.0*npath_table[i][p_ct].c.y,
					     -1.0*npath_table[i][p_ct].c.z);			
			}
			break;
		}
	}
	return;
}

void scene_original_cylinder_model(int mod_ct)
{
	int atm_ct, seq_ct;
	
	for ( seq_ct = my_macro.model[mod_ct].seq_head;
	      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
		if(my_macro.sequence[seq_ct].atom_head == 0)
			continue;	/* Skip empty sequence */
		for (atm_ct = my_macro.sequence[seq_ct].atom_head;
		     atm_ct < my_macro.sequence[seq_ct].atom_tail; atm_ct++) {
			if(atm_ct == my_macro.sequence[seq_ct].atom_head) {
				if(my_macro.atom[atm_ct].atom_name == N_CA){
					glCallList(sphere_30red);
				} else {
					glCallList(sphere_red);
				}
			} else {
				if(my_macro.atom[atm_ct].atom_name == N_CA){
					if (0) glCallList(sphere_30darkblue);
					else glCallList(sphere_white);
				} else {
					if (1) glCallList(sphere_blue); /* blue */
					else glCallList(sphere_white); /* white */
				}
			}
			if (1) draw_cylinder(&avectors[atm_ct], 1015); /* blue */
			else draw_cylinder(&avectors[atm_ct], 1010); /* white */
			glTranslatef(avectors[atm_ct].x, avectors[atm_ct].y,
				     avectors[atm_ct].z);
		}
		if(my_macro.atom[atm_ct].atom_name == N_CA){
			glCallList(sphere_30blue);
		} else {
			glCallList(sphere_blue);
		}
		if(atm_ct < my_macro.atm_num) 
			glTranslatef(avectors[atm_ct].x, avectors[atm_ct].y,
				     avectors[atm_ct].z);
	}
	return;
}

void scene_original_cylinder_model2(int mod_ct)
{
	int atm_ct, seq_ct;
	
	for ( seq_ct = my_macro.model[mod_ct].seq_head;
	      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
		if(my_macro.sequence[seq_ct].atom_head == 0)
			continue;	/* Skip empty sequence */
		for (atm_ct = my_macro.sequence[seq_ct].atom_head;
		     atm_ct < my_macro.sequence[seq_ct].atom_tail; atm_ct++) {
			if(atm_ct == my_macro.sequence[seq_ct].atom_head) {
				if(my_macro.atom[atm_ct].atom_name == N_CA){
					glCallList(sphere_30red);
				} else {
					glCallList(sphere_red);
				}
			} else {
				if(my_macro.atom[atm_ct].atom_name == N_CA){
					if (0) glCallList(sphere_30darkblue);
					else glCallList(sphere_white);
				} else if(my_macro.atom[atm_ct].atom_name == N_CR){
						glCallList(sphere_60red);
				} else if(my_macro.atom[atm_ct].atom_name == N_CY){
					glCallList(sphere_60yellow);
				} else if(my_macro.atom[atm_ct].atom_name == N_CO){
					glCallList(sphere_60orange);	/* NM 2007-04-09 */
				} else if(my_macro.atom[atm_ct].atom_name == N_CB){
					glCallList(sphere_60blue);	/* NM 2007-04-09 */
				} else if(my_macro.atom[atm_ct].atom_name == N_CG){
					glCallList(sphere_60green);	/* NM 2007-04-09 */
				} else if(my_macro.atom[atm_ct].atom_name == N_CU){
					glCallList(sphere_30red);	/* NM 2007-04-09 */
				} else if(my_macro.atom[atm_ct].atom_name == N_CV){
					glCallList(sphere_30orange);	/* NM 2007-04-09 */
				} else if(my_macro.atom[atm_ct].atom_name == N_CW){
					if (1) glCallList(sphere_30darkblue);
					else glCallList(sphere_30white);
				} else {
					if (0) glCallList(sphere_blue); /* blue */
					else glCallList(sphere_white); /* white */
				}
			}
			if (0) draw_cylinder(&avectors[atm_ct], 1015); /* blue */
			else draw_cylinder(&avectors[atm_ct], 1010); /* white */
			glTranslatef(avectors[atm_ct].x, avectors[atm_ct].y,
				     avectors[atm_ct].z);
		}
		if(my_macro.atom[atm_ct].atom_name == N_CA){
			glCallList(sphere_30blue);
		} else {
			glCallList(sphere_blue);
		}
		if(atm_ct < my_macro.atm_num) 
			glTranslatef(avectors[atm_ct].x, avectors[atm_ct].y,
				     avectors[atm_ct].z);
	}
	return;
}

void scene_original(void)	/* Original data */
{
	int mod_ct;

	/* Position of the initial atom */
	glTranslatef(my_macro.atom[1].x, my_macro.atom[1].y,
		     my_macro.atom[1].z);

	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		if (0) scene_original_cylinder_model(mod_ct);
		else scene_original_cylinder_model2(mod_ct);
		break;
	}
	return;
}

void scene_cylinder_model( int mod_ct)
{
	int t_ct, seq_ct;
	
	for (seq_ct = my_macro.model[mod_ct].seq_head;
	     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
		if(my_macro.sequence[seq_ct].tile_head == 0)
			continue;	/* Skip empty sequence */
		for (t_ct = my_macro.sequence[seq_ct].tile_head;
		     t_ct < my_macro.sequence[seq_ct].tile_tail; t_ct++) {
			if (my_macro.atom[slt_tiles[t_ct].error_atom].atom_name
			    == N_CA) {
				glCallList(sphere_30tetra2);
			} else {
				glCallList(sphere_tetra1);
			}
			draw_cylinder(&tvectors[t_ct], 2015);
			glTranslatef(tvectors[t_ct].x, tvectors[t_ct].y,
				     tvectors[t_ct].z);
		}
		if (my_macro.atom[slt_tiles[t_ct].error_atom].atom_name
		    == N_CA) {
			glCallList(sphere_30tetra2);
		} else {
			glCallList(sphere_tetra1);
		}
		if(t_ct < my_macro.tile_num) 
			glTranslatef(tvectors[t_ct].x, tvectors[t_ct].y,
				     tvectors[t_ct].z);
	}	
	return;
}

void scene_solid_model(int mod_ct)
{
	int t_ct, seq_ct;

	for (seq_ct = my_macro.model[mod_ct].seq_head;
	     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
		if(my_macro.sequence[seq_ct].tile_head == 0)
			continue;	/* Skip empty sequence */
		for (t_ct = my_macro.sequence[seq_ct].tile_head;
		     t_ct < my_macro.sequence[seq_ct].tile_tail ; t_ct++) {
			if (my_settings.rotate_tile == PV_TRUE) {
				glPushMatrix();
				glMultMatrixf(slt_tiles[t_ct].rot_mat);
			}
			if (my_macro.atom[slt_tiles[t_ct].error_atom].atom_name
			    == N_CA) {	/* tetra_color1 */
				glCallList(tetra_table2[slt_tiles[t_ct].type]);
			} else {	/* tetra_color2 */
				glCallList(tetra_table[slt_tiles[t_ct].type]);
			}
			if (my_settings.rotate_tile == PV_TRUE) {
				glPopMatrix();
			}
			if (0) draw_cylinder(&tvectors[t_ct], 15);
			glTranslatef(tvectors[t_ct].x, tvectors[t_ct].y,
				     tvectors[t_ct].z);
		}
		if (my_settings.rotate_tile == PV_TRUE) {
			glPushMatrix();
			glMultMatrixf(slt_tiles[t_ct].rot_mat);
		}
		if (my_macro.atom[slt_tiles[t_ct].error_atom].atom_name
		    == N_CA) {
			glCallList(tetra_table2[slt_tiles[t_ct].type]);
		} else {
			glCallList(tetra_table[slt_tiles[t_ct].type]);
		}
		if (my_settings.rotate_tile == PV_TRUE) {
			glPopMatrix();
		}
		if(t_ct < my_macro.tile_num) 
			glTranslatef(tvectors[t_ct].x, tvectors[t_ct].y,
				     tvectors[t_ct].z);
	}
	return;
}

void scene_wire_model(int mod_ct)
{
	int t_ct, seq_ct;
	
	for (seq_ct = my_macro.model[mod_ct].seq_head;
	     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
		if(my_macro.sequence[seq_ct].tile_head == 0)
			continue;	/* Skip empty sequence */
		for (t_ct = my_macro.sequence[seq_ct].tile_head;
		     t_ct < my_macro.sequence[seq_ct].tile_tail; t_ct++) {
			if (my_macro.atom[slt_tiles[t_ct].error_atom].atom_name
			    == N_CA) {	/* tetra_color1 */
				glCallList(tetra_tablew[slt_tiles[t_ct].type]);
			} else {	/* tetra_color2 */
				glCallList(tetra_table2w[slt_tiles[t_ct].type]);
			}
			if (0) draw_cylinder(&tvectors[t_ct], 15);
			glTranslatef(tvectors[t_ct].x, tvectors[t_ct].y,
				     tvectors[t_ct].z);
		}
		if (my_macro.atom[slt_tiles[t_ct].error_atom].atom_name
		    == N_CA) {
			glCallList(tetra_tablew[slt_tiles[t_ct].type]);
		} else {
			glCallList(tetra_table2w[slt_tiles[t_ct].type]);
		}
		if(t_ct < my_macro.tile_num) 
			glTranslatef(tvectors[t_ct].x, tvectors[t_ct].y,
				     tvectors[t_ct].z);
	}
	return;
}

void scene(void)
{
	int mod_ct;

	glTranslatef(slt_tiles[1].cx, slt_tiles[1].cy, slt_tiles[1].cz);

	if (pv_tetra_type <= CYLINDER_TYPE) {
		for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
			scene_cylinder_model(mod_ct);
			break;
		}
	} else if (pv_tetra_type == SOLID_TYPE) {
		for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
			scene_solid_model(mod_ct);
			break;
		}
	} else {
		for (mod_ct=1; mod_ct <= my_macro.mod_num; mod_ct++) {
			scene_wire_model(mod_ct);
			break;
		}
	}
	return;
}

void display(void)
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	glColor3f (1.0, 1.0, 1.0);
	glLoadIdentity ();             /* clear the matrix */
	glTranslatef (CamX, CamY, -15.0);	/* Camera position */
	
	ScaleFactor = 12.0*pv_zoom_in/Ssize[3];
	glScalef (ScaleFactor, ScaleFactor, ScaleFactor);
	
		glPushMatrix();
		glRotatef(Xrot,1,0,0);
		glRotatef(Yrot,0,1,0);
		/* Tcenter[] = The center of the protein */
		glTranslatef(Tcenter[0], Tcenter[1], Tcenter[2]);
		if(SHOW_COOR_AXES) {
			glPushMatrix();
			scene_test();	/* Center mark */
			glPopMatrix();
		}
		if (pv_prt_and_mdl <= BOTH_PRT_AND_MDL) {
			glPushMatrix();
			scene_original();
			glPopMatrix();
		}
		if (pv_prt_and_mdl >= BOTH_PRT_AND_MDL) {
			glPushMatrix();
			scene();
			glPopMatrix();
		}
		glPopMatrix();
	
	glFlush();	
	glutSwapBuffers();
	return;
}

void reshape (int w, int h)
{
	glViewport (0, 0, (GLsizei) w, (GLsizei) h); 
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glFrustum (-1.0, 1.0, -1.0, 1.0, 1.5, 20.0);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glTranslatef(0.0,0.0,-1.0);
	return;
}

void keyboard(unsigned char key, int x, int y)
{
	switch (key) {
		case 27:
			exit(0);
			break;
		case 'B':
			if (pv_blend < 0) {	/* OFF */
				pv_blend = ON;
				glEnable(GL_BLEND);
				glDisable(GL_DEPTH_TEST);
			} else {
				pv_blend = OFF;
				glDisable(GL_BLEND);
				glEnable(GL_DEPTH_TEST);
			}
			break;
		case 'M':
			if (pv_prt_and_mdl < BOTH_PRT_AND_MDL) {
				pv_prt_and_mdl = MODEL_ONLY;
			} else if (pv_prt_and_mdl == BOTH_PRT_AND_MDL) {
				pv_prt_and_mdl = PROTEIN_ONLY;
			} else {
				pv_prt_and_mdl = BOTH_PRT_AND_MDL;
			}
			break;
		case 'S':
			if (pv_tetra_type <= CYLINDER_TYPE) {
				pv_tetra_type = SOLID_TYPE;
			} else if (pv_tetra_type == SOLID_TYPE) {
				pv_tetra_type = WIRE_TYPE;
			} else {
				pv_tetra_type = CYLINDER_TYPE;
			}
			break;
		case 'Z':	/* Zoom in */
			pv_zoom_in = pv_zoom_in * 5/4;
			break;
		case 'z':	/* Zoom out */
			pv_zoom_in = pv_zoom_in * 4/5;
			break;
		case 'R':	/* Rotation */
			if (pv_rot > OFF) {
				pv_rot = OFF;
			} else {
				pv_rot = ON;
				pv_trans = OFF;
			}
			break;
		case 'T':	/* Translation */
			if (pv_trans > OFF) {
				pv_trans = OFF;
			} else {
				pv_trans = ON;
				pv_rot = OFF;
			}
			break;
		case 'C':	/* Undo all */
			Xrot = 0.0;
			Yrot = 0.0;
			CamX = 0.0;
			CamY = 0.0;
			ScaleFactor = 1.0;
			Tcenter[0] = Tcenter_save[0];
			Tcenter[1] = Tcenter_save[1];
			Tcenter[2] = Tcenter_save[2];
			pv_blend = ON;
			pv_tetra_type = SOLID_TYPE;
			pv_zoom_in = 1;
			pv_rot = OFF;
			pv_trans = OFF;
			break;
		case 'Q':	/* NM 2007-04-09 */
			exit(0);
			break;
		default:	/* NM 2007-04-09 */
			break;
	}
	return;
}

/*  -----------------
 *   y-axis
 *   ^
 *   |
 *  (0,0)  ->  x-axis
 *  -----------------
 * (x,y,0): ==> (tmp_a,tmp_b,tmp_c) 
 * glRotatef(-Xrot,1,0,0): ==> (x, cos(-X)y, sin(-X)y)
 * glRotatef(-Yrot,0,1,0): ==> (cos(-Y)x - sin(-Y)sin(-X)y, cos(-X)y,
 *				sin(-Y)x - cos(-Y)sin(-X)y)
 */
static void MouseMove(int x, int y)
{
	if (pv_rot == ON) {
		Yrot = x - WindowWidth/2;
		Xrot = y - WindowHeight/2;
	} else if (pv_trans == ON) {
		CamX = (x - WindowWidth/2) / (100.0*ScaleFactor);
		CamY = -(y - WindowHeight/2)  / (100.0*ScaleFactor);
	}
	glutPostRedisplay();
	return;
}

void idle(void)
{
	glutPostRedisplay();
	return;
}

int main(int argc, char ** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(WindowWidth, WindowHeight);
	glutCreateWindow("Protein Viewer");
	
	myinit (argv, &slt_tiles, &avectors, &tvectors);
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);
	glutPassiveMotionFunc(MouseMove);

	glutMainLoop();

	return 0;
}
